<?php
# api/src/Entity/User.php
namespace App\Entity;
use ApiPlatform\Metadata\ApiProperty;
use ApiPlatform\Metadata\ApiResource;
use ApiPlatform\Metadata\Get;
use ApiPlatform\Metadata\GetCollection;
use ApiPlatform\Metadata\Post;
use ApiPlatform\Metadata\Put;
use App\Controller\CurrentAcountUpdateAction;
use App\Controller\GetCurrentUserAction;
use App\Controller\ResetPasswordAction;
use App\Controller\ResetPasswordForUserAction;
use App\Controller\User\CreateUserRequestAction;
use App\Controller\User\DeleteUserRequestAction;
use App\Controller\User\GetUserRequestAction;
use App\Controller\User\ListUserRequestAction;
use App\Controller\User\UpdateUserRequestAction;
use App\Payload\CurrentAcountUpdate;
use App\Payload\ResetPassword;
use App\Payload\ResetPasswordForUser;
use App\Payload\UserPaylod;
use Doctrine\ORM\Mapping as ORM;
use App\Repository\UserRepository;
use Symfony\Bridge\Doctrine\Validator\Constraints\UniqueEntity;
use Symfony\Component\Security\Core\User\PasswordAuthenticatedUserInterface;
use Symfony\Component\Security\Core\User\UserInterface;
use Symfony\Component\Validator\Constraints as Assert;

#[ApiResource(
    routePrefix: '/auth/user',
    operations: [
        new GetCollection(uriTemplate: '/list',
        controller: ListUserRequestAction::class),
        new Post(uriTemplate: '/create', 
        controller: CreateUserRequestAction::class,
        input: UserPaylod::class),
        new Get(uriTemplate: '/by-id/{id}',
        controller: GetUserRequestAction::class),
        new Put(uriTemplate: '/modify/{id}', 
        controller: UpdateUserRequestAction::class,
        input: UserPaylod::class),
        new Get(uriTemplate: '/delete/{id}',
        controller: DeleteUserRequestAction::class),
        new Get(uriTemplate: '/current-user', 
        controller: GetCurrentUserAction::class),
        new Post(
            uriTemplate: '/reset-password',
            controller: ResetPasswordAction::class,
            input: ResetPassword::class
        ),
        new Post(
            uriTemplate: '/reset-password-for-user',
            controller: ResetPasswordForUserAction::class,
            input: ResetPasswordForUser::class
        ),
        new Post(
            uriTemplate: '/current-update-account',
            controller: CurrentAcountUpdateAction::class,
            input: CurrentAcountUpdate::class
        ),

    ],
    formats: [
        //  'jsonld' => ['application/ld+json'],
        'json' =>    ['application/json'],
        'xml' =>     ['application/xml', 'text/xml'],
    ],
    paginationEnabled: false,
    order: ['createdAt' => 'DESC', 'nom' => 'ASC'],
)]
#[ORM\Entity(repositoryClass: UserRepository::class)]
#[ORM\Table(name: '`user`')]
#[UniqueEntity('email')]
class User  extends Commun implements UserInterface, PasswordAuthenticatedUserInterface
{
    #[ORM\Id]
    #[ORM\Column(type: 'integer')]
    #[ORM\GeneratedValue]
    private ?int $id = null;

    #[Assert\NotBlank]
    #[Assert\Email]
    #[ORM\Column(length: 180, unique: true)]
    private ?string $email = null;

    #[ORM\Column]
    #[ApiProperty(writable: false)]
    private ?string $password = null;

    private ?string $plainPassword = null;

    #[ORM\Column(type: 'json')]
    #[ApiProperty(writable: false)]
    private array $roles = [];

    #[ORM\Column(length: 100, nullable: true)]
    private ?string $nom = null;

    #[ORM\Column(length: 150, nullable: true)]
    private ?string $prenom = null;

    #[ORM\Column(length: 50)]
    private ?string $tel = null;

    #[ApiProperty(writable: false)]
    private ?string $profil = null;

    private ?string $role = null;

    public function __construct()
    {
        
    }

    public function getId(): ?int
    {
        return $this->id;
    }

    public function getEmail(): ?string
    {
        return $this->email;
    }

    public function setEmail(string $email): self
    {
        $this->email = $email;
        return $this;
    }

    /**
     * @see PasswordAuthenticatedUserInterface
     */
    public function getPassword(): string
    {
        return $this->password;
    }

    public function setPassword(string $password): self
    {
        $this->password = $password;
        return $this;
    }

    public function getPlainPassword(): ?string
    {
        return $this->plainPassword;
    }

    public function setPlainPassword(?string $plainPassword): self
    {
        $this->plainPassword = $plainPassword;
        return $this;
    }

    /**
     * @see UserInterface
     */
    public function getRoles(): array
    {
        $roles = $this->roles;
        //$roles[] = 'ROLE_USER';
        return array_unique($roles);
    }

    public function setRoles(array $roles): self
    {
        $this->roles = $roles;
        return $this;
    }

    /**
     * A visual identifier that represents this user.
     *
     * @see UserInterface
     */
    public function getUserIdentifier(): string
    {
        return (string) $this->email;
    }

    /**
     * @see UserInterface
     */
    public function eraseCredentials(): void
    {
        $this->plainPassword = null;
    }

    public function getNom(): ?string
    {
        return $this->nom;
    }

    public function setNom(?string $nom): static
    {
        $this->nom = $nom;

        return $this;
    }

    public function getPrenom(): ?string
    {
        return $this->prenom;
    }

    public function setPrenom(?string $prenom): static
    {
        $this->prenom = $prenom;

        return $this;
    }

    public function getTel(): ?string
    {
        return $this->tel;
    }

    public function setTel(string $tel): static
    {
        $this->tel = $tel;

        return $this;
    }

    public function getProfil(): ?string
    {
        switch ($this->getRoles()[0]) {
            case 'ROLE_USER':
                $this->profil = "Utilisateur";
                break;
            case 'ROLE_ADMIN':
                $this->profil = "Administrateur";
                break;
            case 'ROLE_SUPER_ADMIN':
                $this->profil = "Super Administrateur";
                break;
        }
        return $this->profil;
    }

    public function setProfil(): self
    {
        switch ($this->getRoles()[0]) {
            case 'ROLE_USER':
                $this->profil = "Utilisateur";
                break;
            case 'ROLE_ADMIN':
                $this->profil = "Administrateur";
                break;
            case 'ROLE_SUPER_ADMIN':
                $this->profil = "Super Administrateur";
                break;
        }
        return $this;
    }

    public function getRole(): ?string
    {
        return $this->role;
    }

    public function setRole(string $role): self
    {
        $this->role = $role;
        return $this;
    }
}
